from typing import List

from langchain.schema import ChatMessage


def get_buffer_string(messages: List[ChatMessage]) -> str:
    """Get buffer string of messages."""
    return "\n".join([f"{gen.role}: {gen.text}" for gen in messages])



from abc import ABC
from typing import Dict

from pydantic import root_validator

from langchain.chains.base import Chain
from langchain.memory.chat_memory import ChatMemory


class BaseChatChain(Chain, ABC):
    human_prefix: str = "user"
    ai_prefix: str = "assistant"

    @root_validator()
    def validate_memory_keys(cls, values: Dict) -> Dict:
        """Validate that the human and ai prefixes line up."""
        if "memory" in values:
            memory = values["memory"]
            if isinstance(memory, ChatMemory):
                if memory.human_prefix != values["human_prefix"]:
                    raise ValueError(
                        f"Memory human_prefix ({memory.human_prefix}) must "
                        f"match chain human_prefix ({values['human_prefix']})"
                    )
                if memory.ai_prefix != values["ai_prefix"]:
                    raise ValueError(
                        f"Memory ai_prefix ({memory.ai_prefix}) must "
                        f"match chain ai_prefix ({values['ai_prefix']})"
                    )
        return values



"""Loader that loads GitBook."""
from typing import Any, List, Optional

from langchain.docstore.document import Document
from langchain.document_loaders.web_base import WebBaseLoader


class GitbookLoader(WebBaseLoader):
    """Load GitBook data.

    1. load from either a single page, or
    2. load all (relative) paths in the navbar.
    """

    def __init__(self, web_page: str, load_all_paths: bool = False):
        """Initialize with web page and whether to load all paths."""
        super().__init__(web_page)
        self.load_all_paths = load_all_paths

    def load(self) -> List[Document]:
        """Fetch text from one single GitBook page."""
        if self.load_all_paths:
            soup_info = self.scrape()
            relative_paths = self._get_paths(soup_info)
            documents = []
            for path in relative_paths:
                url = self.web_path + path
                print(f"Fetching text from {url}")
                soup_info = self._scrape(url)
                documents.append(self._get_document(soup_info, url))
            return documents
        else:
            soup_info = self.scrape()
            return [self._get_document(soup_info, self.web_path)]

    def _get_document(self, soup: Any, custom_url: Optional[str] = None) -> Document:
        """Fetch content from page and return Document."""
        page_content_raw = soup.find("main")
        content = page_content_raw.get_text(separator="\n").strip()
        title_if_exists = page_content_raw.find("h1")
        title = title_if_exists.text if title_if_exists else ""
        metadata = {
            "source": custom_url if custom_url else self.web_path,
            "title": title,
        }
        return Document(page_content=content, metadata=metadata)

    def _get_paths(self, soup: Any) -> List[str]:
        """Fetch all relative paths in the navbar."""
        nav = soup.find("nav")
        links = nav.findAll("a")
        # only return relative links
        return [link.get("href") for link in links if link.get("href")[0] == "/"]



"""Chain that carries on a conversation and calls an LLM."""
from __future__ import annotations

from typing import Any, Dict, List

from pydantic import BaseModel, Extra, Field

from langchain.chains.conversation.prompt import PROMPT
from langchain.chat.base import BaseChatChain
from langchain.chat.memory import SimpleChatMemory
from langchain.chat_models.base import BaseChat
from langchain.prompts.base import BasePromptTemplate
from langchain.schema import ChatMessage


class ConversationChain(BaseChatChain, BaseModel):
    """Chain to have a conversation and load context from memory.

    Example:
        .. code-block:: python

            from langchain import ConversationChain, OpenAI
            conversation = ConversationChain(llm=OpenAI())
    """

    model: BaseChat
    memory: SimpleChatMemory = Field(default_factory=SimpleChatMemory)
    """Default memory store."""
    prompt: BasePromptTemplate = PROMPT
    """Default conversation prompt to use."""
    input_key: str = "input"  #: :meta private:
    output_key: str = "response"  #: :meta private:
    starter_messages: List[ChatMessage] = Field(default_factory=list)

    @classmethod
    def from_model(cls, model: BaseModel, **kwargs: Any) -> ConversationChain:
        """From model. Future proofing."""
        return cls(model=model, **kwargs)

    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        new_message = ChatMessage(text=inputs[self.input_key], role=self.human_prefix)
        messages = self.starter_messages + self.memory.messages + [new_message]
        output = self.model.run(messages)
        return {self.output_key: output.text}

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Use this since so some prompt vars come from history."""
        return [self.input_key]



"""Memory modules for conversation prompts."""
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, root_validator

from langchain.chains.base import Memory
from langchain.chains.conversation.prompt import (
    ENTITY_EXTRACTION_PROMPT,
    ENTITY_SUMMARIZATION_PROMPT,
    KNOWLEDGE_TRIPLE_EXTRACTION_PROMPT,
    SUMMARY_PROMPT,
)
from langchain.chains.llm import LLMChain
from langchain.graphs.networkx_graph import (
    NetworkxEntityGraph,
    get_entities,
    parse_triples,
)
from langchain.llms.base import BaseLLM
from langchain.memory.chat_memory import ChatMemory
from langchain.memory.utils import get_buffer_string
from langchain.prompts.base import BasePromptTemplate
from langchain.schema import ChatMessage


def _get_prompt_input_key(inputs: Dict[str, Any], memory_variables: List[str]) -> str:
    # "stop" is a special key that can be passed as input but is not used to
    # format the prompt.
    prompt_input_keys = list(set(inputs).difference(memory_variables + ["stop"]))
    if len(prompt_input_keys) != 1:
        raise ValueError(f"One input key expected got {prompt_input_keys}")
    return prompt_input_keys[0]


class CombinedMemory(Memory, BaseModel):
    """Class for combining multiple memories' data together."""

    memories: List[Memory]
    """For tracking all the memories that should be accessed."""

    @property
    def memory_variables(self) -> List[str]:
        """All the memory variables that this instance provides."""
        """Collected from the all the linked memories."""

        memory_variables = []

        for memory in self.memories:
            memory_variables.extend(memory.memory_variables)

        return memory_variables

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        """Load all vars from sub-memories."""
        memory_data: Dict[str, Any] = {}

        # Collect vars from all sub-memories
        for memory in self.memories:
            data = memory.load_memory_variables(inputs)
            memory_data = {
                **memory_data,
                **data,
            }

        return memory_data

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this session for every memory."""
        # Save context for all sub-memories
        for memory in self.memories:
            memory.save_context(inputs, outputs)

    def clear(self) -> None:
        """Clear context from this session for every memory."""
        for memory in self.memories:
            memory.clear()


class ChatMemoryMixin(Memory):
    chat_memory: ChatMemory
    human_prefix: str = "Human"
    ai_prefix: str = "AI"
    output_key: Optional[str] = None
    input_key: Optional[str] = None

    @root_validator(pre=True)
    def add_chat_memory(cls, values: Dict) -> Dict:
        """Add chat memory data structure."""
        human_prefix = values.get("human_prefix", "Human")
        ai_prefix = values.get("ai_prefix", "AI")
        values["chat_memory"] = ChatMemory(
            human_prefix=human_prefix, ai_prefix=ai_prefix
        )
        return values

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this conversation to buffer."""
        if self.input_key is None:
            prompt_input_key = _get_prompt_input_key(inputs, self.memory_variables)
        else:
            prompt_input_key = self.input_key
        if self.output_key is None:
            if len(outputs) != 1:
                raise ValueError(f"One output key expected, got {outputs.keys()}")
            output_key = list(outputs.keys())[0]
        else:
            output_key = self.output_key
        self.chat_memory.add_user_message(inputs[prompt_input_key])
        self.chat_memory.add_ai_message(outputs[output_key])

    def clear(self) -> None:
        """Clear memory contents."""
        self.chat_memory.clear()


class ConversationBufferMemory(ChatMemoryMixin, BaseModel):
    """Buffer for storing conversation memory."""

    memory_key: str = "history"  #: :meta private:

    @property
    def buffer(self) -> str:
        """String buffer of memory."""
        return get_buffer_string(self.chat_memory.messages)

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        """Return history buffer."""
        return {self.memory_key: self.buffer}


class ConversationBufferWindowMemory(ChatMemoryMixin, BaseModel):
    """Buffer for storing conversation memory."""

    memory_key: str = "history"  #: :meta private:
    k: int = 5

    @property
    def buffer(self) -> List[ChatMessage]:
        """String buffer of memory."""
        return self.chat_memory.messages

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        """Return history buffer."""
        return {self.memory_key: get_buffer_string(self.buffer[-self.k * 2 :])}


# For legacy naming reasons
ConversationalBufferWindowMemory = ConversationBufferWindowMemory


class ConversationSummaryMemory(ChatMemoryMixin, BaseModel):
    """Conversation summarizer to memory."""

    buffer: str = ""
    """Prefix to use for AI generated responses."""
    llm: BaseLLM
    prompt: BasePromptTemplate = SUMMARY_PROMPT
    memory_key: str = "history"  #: :meta private:

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        """Return history buffer."""
        return {self.memory_key: self.buffer}

    @root_validator()
    def validate_prompt_input_variables(cls, values: Dict) -> Dict:
        """Validate that prompt input variables are consistent."""
        prompt_variables = values["prompt"].input_variables
        expected_keys = {"summary", "new_lines"}
        if expected_keys != set(prompt_variables):
            raise ValueError(
                "Got unexpected prompt input variables. The prompt expects "
                f"{prompt_variables}, but it should have {expected_keys}."
            )
        return values

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this conversation to buffer."""
        super().save_context(inputs, outputs)
        new_lines = get_buffer_string(self.chat_memory.messages[-2:])
        chain = LLMChain(llm=self.llm, prompt=self.prompt)
        self.buffer = chain.predict(summary=self.buffer, new_lines=new_lines)

    def clear(self) -> None:
        """Clear memory contents."""
        super().clear()
        self.buffer = ""


class ConversationEntityMemory(ChatMemoryMixin, BaseModel):
    """Entity extractor & summarizer to memory."""

    """Prefix to use for AI generated responses."""
    llm: BaseLLM
    entity_extraction_prompt: BasePromptTemplate = ENTITY_EXTRACTION_PROMPT
    entity_summarization_prompt: BasePromptTemplate = ENTITY_SUMMARIZATION_PROMPT
    store: Dict[str, Optional[str]] = {}
    entity_cache: List[str] = []
    k: int = 3
    chat_history_key: str = "history"

    @property
    def buffer(self) -> List[ChatMessage]:
        return self.chat_memory.messages

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return ["entities", self.chat_history_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Return history buffer."""
        chain = LLMChain(llm=self.llm, prompt=self.entity_extraction_prompt)
        if self.input_key is None:
            prompt_input_key = _get_prompt_input_key(inputs, self.memory_variables)
        else:
            prompt_input_key = self.input_key
        output = chain.predict(
            history=get_buffer_string(self.buffer[-self.k * 2 :]),
            input=inputs[prompt_input_key],
        )
        if output.strip() == "NONE":
            entities = []
        else:
            entities = [w.strip() for w in output.split(",")]
        entity_summaries = {}
        for entity in entities:
            entity_summaries[entity] = self.store.get(entity, "")
        self.entity_cache = entities
        return {
            self.chat_history_key: get_buffer_string(self.buffer[-self.k * 2 :]),
            "entities": entity_summaries,
        }

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this conversation to buffer."""
        super().save_context(inputs, outputs)
        if self.input_key is None:
            prompt_input_key = _get_prompt_input_key(inputs, self.memory_variables)
        else:
            prompt_input_key = self.input_key
        for entity in self.entity_cache:
            chain = LLMChain(llm=self.llm, prompt=self.entity_summarization_prompt)
            # key value store for entity
            existing_summary = self.store.get(entity, "")
            output = chain.predict(
                summary=existing_summary,
                history=get_buffer_string(self.buffer[-self.k * 2 :]),
                input=inputs[prompt_input_key],
                entity=entity,
            )
            self.store[entity] = output.strip()

    def clear(self) -> None:
        """Clear memory contents."""
        self.chat_memory.clear()
        self.store = {}


class ConversationSummaryBufferMemory(ChatMemoryMixin, BaseModel):
    """Buffer with summarizer for storing conversation memory."""

    max_token_limit: int = 2000
    moving_summary_buffer: str = ""
    llm: BaseLLM
    prompt: BasePromptTemplate = SUMMARY_PROMPT
    memory_key: str = "history"

    @property
    def buffer(self) -> List[ChatMessage]:
        return self.chat_memory.messages

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        """Return history buffer."""
        if self.moving_summary_buffer == "":
            return {self.memory_key: get_buffer_string(self.buffer)}
        memory_val = self.moving_summary_buffer + "\n" + get_buffer_string(self.buffer)
        return {self.memory_key: memory_val}

    @root_validator()
    def validate_prompt_input_variables(cls, values: Dict) -> Dict:
        """Validate that prompt input variables are consistent."""
        prompt_variables = values["prompt"].input_variables
        expected_keys = {"summary", "new_lines"}
        if expected_keys != set(prompt_variables):
            raise ValueError(
                "Got unexpected prompt input variables. The prompt expects "
                f"{prompt_variables}, but it should have {expected_keys}."
            )
        return values

    def get_num_tokens_list(self, arr: List[ChatMessage]) -> List[int]:
        """Get list of number of tokens in each string in the input array."""
        return [self.llm.get_num_tokens(get_buffer_string([x])) for x in arr]

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this conversation to buffer."""
        super().save_context(inputs, outputs)
        # Prune buffer if it exceeds max token limit
        buffer = self.chat_memory.messages
        curr_buffer_length = sum(self.get_num_tokens_list(buffer))
        if curr_buffer_length > self.max_token_limit:
            pruned_memory = []
            while curr_buffer_length > self.max_token_limit:
                pruned_memory.append(buffer.pop(0))
                curr_buffer_length = sum(self.get_num_tokens_list(buffer))
            chain = LLMChain(llm=self.llm, prompt=self.prompt)
            self.moving_summary_buffer = chain.predict(
                summary=self.moving_summary_buffer,
                new_lines=(get_buffer_string(pruned_memory)),
            )

    def clear(self) -> None:
        """Clear memory contents."""
        super().clear()
        self.moving_summary_buffer = ""


class ConversationKGMemory(ChatMemoryMixin, BaseModel):
    """Knowledge graph memory for storing conversation memory.

    Integrates with external knowledge graph to store and retrieve
    information about knowledge triples in the conversation.
    """

    k: int = 2
    kg: NetworkxEntityGraph = Field(default_factory=NetworkxEntityGraph)
    knowledge_extraction_prompt: BasePromptTemplate = KNOWLEDGE_TRIPLE_EXTRACTION_PROMPT
    entity_extraction_prompt: BasePromptTemplate = ENTITY_EXTRACTION_PROMPT
    llm: BaseLLM
    """Number of previous utterances to include in the context."""
    memory_key: str = "history"  #: :meta private:

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Return history buffer."""
        entities = self._get_current_entities(inputs)
        summaries = {}
        for entity in entities:
            knowledge = self.kg.get_entity_knowledge(entity)
            if knowledge:
                summaries[entity] = ". ".join(knowledge) + "."
        if summaries:
            summary_strings = [
                f"On {entity}: {summary}" for entity, summary in summaries.items()
            ]
            context_str = "\n".join(summary_strings)
        else:
            context_str = ""
        return {self.memory_key: context_str}

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    def _get_prompt_input_key(self, inputs: Dict[str, Any]) -> str:
        """Get the input key for the prompt."""
        if self.input_key is None:
            return _get_prompt_input_key(inputs, self.memory_variables)
        return self.input_key

    def _get_prompt_output_key(self, outputs: Dict[str, Any]) -> str:
        """Get the output key for the prompt."""
        if self.output_key is None:
            if len(outputs) != 1:
                raise ValueError(f"One output key expected, got {outputs.keys()}")
            return list(outputs.keys())[0]
        return self.output_key

    def _get_current_entities(self, inputs: Dict[str, Any]) -> List[str]:
        """Get the current entities in the conversation."""
        prompt_input_key = self._get_prompt_input_key(inputs)
        chain = LLMChain(llm=self.llm, prompt=self.entity_extraction_prompt)
        output = chain.predict(
            history=get_buffer_string(self.chat_memory.messages[-self.k :]),
            input=inputs[prompt_input_key],
        )
        return get_entities(output)

    def _get_and_update_kg(self, inputs: Dict[str, Any]) -> None:
        """Get and update knowledge graph from the conversation history."""
        chain = LLMChain(llm=self.llm, prompt=self.knowledge_extraction_prompt)
        prompt_input_key = self._get_prompt_input_key(inputs)
        output = chain.predict(
            history=get_buffer_string(self.chat_memory.messages[-self.k :]),
            input=inputs[prompt_input_key],
            verbose=True,
        )
        knowledge = parse_triples(output)
        for triple in knowledge:
            self.kg.add_triple(triple)

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this conversation to buffer."""
        super().save_context(inputs, outputs)
        self._get_and_update_kg(inputs)

    def clear(self) -> None:
        """Clear memory contents."""
        super().clear()
        self.kg.clear()



"""OpenAI chat wrapper."""
import logging
from typing import Any, Callable, Dict, List, Mapping, Optional

from pydantic import BaseModel, Extra, Field, root_validator
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from langchain.chat_models.base import BaseChat
from langchain.schema import ChatGeneration, ChatMessage, ChatResult
from langchain.utils import get_from_dict_or_env

logger = logging.getLogger(__file__)


class OpenAI(BaseChat, BaseModel):
    """Wrapper around OpenAI Chat large language models.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.

    Any parameters that are valid to be passed to the openai.create call can be passed
    in, even if not explicitly saved on this class.

    Example:
        .. code-block:: python

            from langchain.chat import OpenAI
            openai = OpenAI(model_name="gpt-3.5-turbo")
    """

    client: Any  #: :meta private:
    model_name: str = "gpt-3.5-turbo"
    """Model name to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    openai_api_key: Optional[str] = None
    max_retries: int = 6
    """Maximum number of retries to make when generating."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    @root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = {field.alias for field in cls.__fields__.values()}

        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name not in all_required_field_names:
                if field_name in extra:
                    raise ValueError(f"Found {field_name} supplied twice.")
                extra[field_name] = values.pop(field_name)
        values["model_kwargs"] = extra
        return values

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        openai_api_key = get_from_dict_or_env(
            values, "openai_api_key", "OPENAI_API_KEY"
        )
        try:
            import openai

            openai.api_key = openai_api_key
        except ImportError:
            raise ValueError(
                "Could not import openai python package. "
                "Please it install it with `pip install openai`."
            )
        try:
            values["client"] = openai.ChatCompletion
        except AttributeError:
            raise ValueError(
                "`openai` has no `ChatCompletion` attribute, this is likely "
                "due to an old version of the openai package. Try upgrading it "
                "with `pip install --upgrade openai`."
            )
        return values

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        return self.model_kwargs

    def _create_retry_decorator(self) -> Callable[[Any], Any]:
        import openai

        min_seconds = 4
        max_seconds = 10
        # Wait 2^x * 1 second between each retry starting with
        # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
        return retry(
            reraise=True,
            stop=stop_after_attempt(self.max_retries),
            wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
            retry=(
                retry_if_exception_type(openai.error.Timeout)
                | retry_if_exception_type(openai.error.APIError)
                | retry_if_exception_type(openai.error.APIConnectionError)
                | retry_if_exception_type(openai.error.RateLimitError)
                | retry_if_exception_type(openai.error.ServiceUnavailableError)
            ),
            before_sleep=before_sleep_log(logger, logging.WARNING),
        )

    def completion_with_retry(self, **kwargs: Any) -> Any:
        """Use tenacity to retry the completion call."""
        retry_decorator = self._create_retry_decorator()

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            return self.client.create(**kwargs)

        return _completion_with_retry(**kwargs)

    def _generate(
        self, messages: List[ChatMessage], stop: Optional[List[str]] = None
    ) -> ChatResult:
        params: Dict[str, Any] = {**{"model": self.model_name}, **self._default_params}
        if stop is not None:
            if "stop" in params:
                raise ValueError("`stop` found in both the input and default params.")
            params["stop"] = stop
        message_dicts = [{"role": m.role, "content": m.text} for m in messages]
        response = self.completion_with_retry(messages=message_dicts, **params)
        generations = []
        for res in response["choices"]:
            message = ChatMessage(
                text=res["message"]["content"], role=res["message"]["role"]
            )
            gen = ChatGeneration(message=message)
            generations.append(gen)
        return ChatResult(generations=generations)

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {**{"model_name": self.model_name}, **self._default_params}

    @property
    def _llm_type(self) -> str:
        """Return type of chat."""
        return "openai"



from abc import ABC, abstractmethod
from typing import List, Optional

from langchain.schema import ChatGeneration, ChatMessage, ChatResult


class BaseChat(ABC):
    def generate(
        self, messages: List[ChatMessage], stop: Optional[List[str]] = None
    ) -> ChatResult:
        """Top Level call"""
        # Nothing here now, but future proofing.
        return self._generate(messages, stop=stop)

    @abstractmethod
    def _generate(
        self, messages: List[ChatMessage], stop: Optional[List[str]] = None
    ) -> ChatResult:
        """Top Level call"""

    def run(
        self, messages: List[ChatMessage], stop: Optional[List[str]] = None
    ) -> ChatMessage:
        res = self.generate(messages, stop=stop)
        return res.generations[0].message


class SimpleChat(BaseChat):
    role: str = "assistant"

    def _generate(
        self, messages: List[ChatMessage], stop: Optional[List[str]] = None
    ) -> ChatResult:
        output_str = self._call(messages, stop=stop)
        message = ChatMessage(text=output_str, role=self.role)
        generation = ChatGeneration(message=message)
        return ChatResult(generations=[generation])

    @abstractmethod
    def _call(
        self, messages: List[ChatMessage], stop: Optional[List[str]] = None
    ) -> str:
        """Simpler interface."""



"""Question Answering."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Extra, Field

from langchain.chains.conversation.prompt import PROMPT
from langchain.chat.base import BaseChatChain
from langchain.chat_models.base import BaseChat
from langchain.prompts.base import BasePromptTemplate
from langchain.schema import ChatMessage


def _get_default_starter_messages() -> List[ChatMessage]:
    prompt = (
        "You are chatbot optimized for question answering. "
        "Your job is to answer the most recent user question based "
        "ONLY on the information they have told you before. "
        "Do NOT use other information than what is provided in previous "
        "messages by the human."
    )
    return [ChatMessage(text=prompt, role="system")]


class QAChain(BaseChatChain, BaseModel):
    """Chain to have a conversation and load context from memory.

    Example:
        .. code-block:: python

            from langchain import ConversationChain, OpenAI
            conversation = ConversationChain(llm=OpenAI())
    """

    model: BaseChat
    """Default memory store."""
    prompt: BasePromptTemplate = PROMPT
    """Default conversation prompt to use."""
    question_key: str = "question"  #: :meta private:
    documents_key: str = "input_documents"  #: :meta private:
    output_key: str = "response"  #: :meta private:
    starter_messages: List[ChatMessage] = Field(default_factory=list)

    @classmethod
    def from_model(
        cls,
        model: BaseChat,
        starter_messages: Optional[List[ChatMessage]] = None,
        **kwargs: Any,
    ) -> QAChain:
        """From model. Future proofing."""
        if starter_messages is not None:
            _starter_messages = starter_messages
        else:
            _starter_messages = _get_default_starter_messages()
        return cls(model=model, **kwargs)

    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        new_message = ChatMessage(
            text=inputs[self.question_key], role=self.human_prefix
        )
        docs = inputs[self.documents_key]
        doc_messages = [
            ChatMessage(text=doc.page_content, role=self.human_prefix) for doc in docs
        ]
        messages = self.starter_messages + doc_messages + [new_message]
        output = self.model.run(messages)
        return {self.output_key: output.text}

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Use this since so some prompt vars come from history."""
        return [self.question_key, self.documents_key]



from typing import Any, Dict, List, Optional

from langchain.chains.base import Memory
from langchain.memory.chat_memory import ChatMemory


def _get_prompt_input_key(inputs: Dict[str, Any], key: Optional[str]) -> str:
    if key is not None:
        return key
    # "stop" is a special key that can be passed as input but is not used to
    # format the prompt.
    prompt_input_keys = list(set(inputs).difference(["stop"]))
    if len(prompt_input_keys) != 1:
        raise ValueError(f"One input key expected got {prompt_input_keys}")
    return prompt_input_keys[0]


class SimpleChatMemory(Memory, ChatMemory):
    input_key: Optional[str] = None
    output_key: Optional[str] = None

    def clear(self) -> None:
        self.clear()

    @property
    def memory_variables(self) -> List[str]:
        return []

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        return {}

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        self.add_user_message(inputs[_get_prompt_input_key(inputs, self.input_key)])
        self.add_ai_message(outputs[_get_prompt_input_key(outputs, self.output_key)])


class ChatHistoryMemory(SimpleChatMemory):
    chat_history_key: str = "chat_history"

    @property
    def memory_variables(self) -> List[str]:
        return [self.chat_history_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        return {self.chat_history_key: self.messages}



"""Chain for chatting with a vector database."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from langchain.chains.chat_vector_db.prompts import CONDENSE_QUESTION_PROMPT
from langchain.chains.llm import LLMChain
from langchain.chat.base import BaseChatChain
from langchain.chat.question_answering import QAChain
from langchain.chat_models.base import BaseChat
from langchain.llms.base import BaseLLM
from langchain.memory.utils import get_buffer_string
from langchain.prompts.base import BasePromptTemplate
from langchain.schema import ChatMessage
from langchain.vectorstores.base import VectorStore


class ChatVectorDBChain(BaseChatChain, BaseModel):
    """Chain for chatting with a vector database."""

    vectorstore: VectorStore
    qa_chain: QAChain
    question_generator: LLMChain
    output_key: str = "answer"
    return_source_documents: bool = False
    top_k_docs_for_context: int = 4
    """Return the source documents."""

    @property
    def _chain_type(self) -> str:
        return "chat-vector-db"

    @property
    def input_keys(self) -> List[str]:
        """Input keys."""
        return ["question", "chat_history"]

    @property
    def output_keys(self) -> List[str]:
        """Return the output keys.

        :meta private:
        """
        _output_keys = [self.output_key]
        if self.return_source_documents:
            _output_keys = _output_keys + ["source_documents"]
        return _output_keys

    @classmethod
    def from_llm(
        cls,
        *,
        llm: BaseLLM,
        model: BaseChat,
        vectorstore: VectorStore,
        starter_messages: Optional[List[ChatMessage]] = None,
        condense_question_prompt: BasePromptTemplate = CONDENSE_QUESTION_PROMPT,
        **kwargs: Any,
    ) -> ChatVectorDBChain:
        """Load chain from LLM."""
        qa_chain = QAChain.from_model(model, starter_messages=starter_messages)
        condense_question_chain = LLMChain(llm=llm, prompt=condense_question_prompt)
        return cls(
            vectorstore=vectorstore,
            qa_chain=qa_chain,
            question_generator=condense_question_chain,
            **kwargs,
        )

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        question = inputs["question"]
        chat_history_str = get_buffer_string(inputs["chat_history"])
        vectordbkwargs = inputs.get("vectordbkwargs", {})
        if chat_history_str:
            new_question = self.question_generator.run(
                question=question, chat_history=chat_history_str
            )
        else:
            new_question = question
        docs = self.vectorstore.similarity_search(
            new_question, k=self.top_k_docs_for_context, **vectordbkwargs
        )
        args = {
            self.qa_chain.documents_key: docs,
            self.qa_chain.question_key: new_question,
        }

        result = self.qa_chain(args)
        answer = result[self.qa_chain.output_key]
        if self.return_source_documents:
            return {self.output_key: answer, "source_documents": docs}
        else:
            return {self.output_key: answer}



from typing import List

from pydantic import BaseModel, Field

from langchain.schema import ChatMessage


class ChatMemory(BaseModel):
    human_prefix: str = "user"
    ai_prefix: str = "assistant"
    messages: List[ChatMessage] = Field(default_factory=list)

    def add_user_message(self, message: str) -> None:
        gen = ChatMessage(text=message, role=self.human_prefix)
        self.messages.append(gen)

    def add_ai_message(self, message: str) -> None:
        gen = ChatMessage(text=message, role=self.ai_prefix)
        self.messages.append(gen)

    def clear(self) -> None:
        self.messages = []
